/******************************************************************************
 *
 * Copyright (c) 2019, the Perspective Authors.
 *
 * This file is part of the Perspective library, distributed under the terms of
 * the Apache License 2.0.  The full license can be found in the LICENSE file.
 *
 */
#include <perspective/first.h>
#include <perspective/base.h>
#include <perspective/binding.h>
using namespace perspective;
namespace perspective {
namespace binding {}
} // namespace perspective
